#!/corp/global/tools/bin/python3.4
import re
import os
import sys
import pexpect
import xml.etree.cElementTree as ET
from pexpect import pxssh
import pprint
import getpass
import argparse
import unittest
from time import gmtime, strftime
from functools import reduce
from io import BytesIO
from subprocess import Popen, PIPE, call
import xml.dom.minidom
import itertools

ssh_session = ''
conf_params = ['DEVICEIP', 'DEVICEUNAME', 'DEVICEPASSWD', 'NETCONFUNAME', 'PASSWD', 'OUTPUTDIR', 'CONFIGFILE']
param_dict = {}
keys_dict = {}
grouped_keys = []
input_keys = []

sub_tree_node = None
gen_nc_dump = False
uri_list = []
outfile =''
ret_code = ''
query = ''
hello = '''<?xml version="1.0" encoding="UTF-8"?>
<hello xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
  <capabilities>
    <capability>urn:ietf:params:netconf:base:1.0</capability>
  </capabilities>
</hello>
]]>]]>
'''

close = '''<?xml version="1.0" encoding="UTF-8"?>
<rpc xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="2">
  <close-session/>
</rpc>
'''
base_path = os.path.dirname(os.path.realpath(__file__))

def create_net_conf_tree():
# bin/pyang_1 -o finale -f tree -p yang/:fabos/src/ fabos/src/yang/*oper
    
    print("\nGenerating flattened YANG tree from the operational YANG FILES")
    print("*********************************************************************************************\n")
    os.system("python2.7 "+base_path+"/../../../../../../build/build/target26/confd/LATEST/x86_64/bin/pyang -o rest_tree -f tree -p yang/:"+base_path+"/../../../../../src/ "+base_path+"/../../../../../src/yang/*oper*.yang >> log_file 2>&1")
    print("*********************************************************************************************\n")

def doSsh(user, password, sport, host, config_file):
    child = pexpect.spawn('ssh '+user+'@'+host)
    child.expect(".*assword:")
    child.send(password+"\r")
    #child.expect('.*to proceed.')
    #child.send('\003')
    child.expect('>')
    child.sendline('noscli')
    child.expect('#')
    child.sendline('conf t\r\r\r\r')
    child.expect('#')
    child.sendline('\r'.join([line.strip() for line in open(config_file, 'r')])+'\r')


def init_net_conf_ssh(user, password, sport, host, options):
    global ssh_session
    try:
        ssh_session = pexpect.spawn('ssh '+user+'@'+host+" -o \"StrictHostKeyChecking no\" -p "+sport+options)
        ssh_session.expect(".*assword:")
        ssh_session.send(password+"\r")
        ssh_session.expect(']]>]]>')
        ssh_session.sendline(hello)
    except pexpect.ExceptionPexpect as e:
        print("Failed to initialize NetConf Session.\n")
        print("Reasons for failure could be:\n")
        print("1. DUT is not reachable.\n")
        print("2. Host key verification failed.\n")
        exit()

def close_net_conf_ssh():
    ssh_session.sendline(close)

def read_conf_file():
    global conf_params
    global param_dict

    if os.path.exists(os.getcwd()+"/ncBreezerc"):
        for line in open("ncBreezerc", "r"):
            if len(line.strip()):
                if line.startswith("#"):
                    pass
                else:
                    [key, val] = line.strip().split("=")
                    param_dict[key] = val
        #param_dict = dict([line.strip().split("=") for line in open("rBreezerc", "r")])
    else:
        conf = open("rBreezerc", "w")
        for param in conf_params:
            val = input(param+"[Empty]=")
            conf.write(param+"="+val+"\n")
            param_dict[param] = val
        conf.close()

def update_keys(key, val):
    global keys_dict
    if key in keys_dict.keys():
        if val not in keys_dict[key]:
            keys_dict[key].append(val)

def process_keys(key_val_list):
   keys_list = re.findall("_(.[^_]*)_", sub_tree_node)

   if len(keys_list) != len(key_val_list):
       print("Error: Please enter exactly the same number of key values as there are keys in the URI")
       exit()

   for key, val in zip(keys_list, key_val_list):   
       update_keys(key, val) 

def optional_args():
    global dest_dir
    global sub_tree_node
    global dump_gen
    global input_keys

    parser = argparse.ArgumentParser(description='Run Operational commands test suite')
    parser.add_argument('-t', dest="sub_tree_node", type=str, action='store',
                   help='Subtree to start the suite')
    parser.add_argument('-g', dest="dump_gen", action='store_true',
                   help='Generate NETCONF operational URI dump')
    parser.add_argument('-k', dest="keys", type=str, nargs = "+", action='store',
                   help='Key values for the Keys in the URIs')

    args = parser.parse_args()

    sub_tree_node = args.sub_tree_node or None
    if sub_tree_node:
        if "v@l_" in sub_tree_node:
            input_keys = args.keys
    dump_gen = args.dump_gen

    if sub_tree_node and dump_gen:
        print("Error: Options '-g' and '-t' cannot coexist")
        exit()

    print("\n*********************************************************************************************\n")
    if sub_tree_node is not None:
        print("Starting from sub tree "+sub_tree_node)


class read_tree:
    global uri_list
    global keys_dict

    def _init_(self):
        print("Inited Class")

    def get_level(self, len):
        return (len-8)//3

    def print_uri(self):
        cnt = 0
        list_elem = ['']
        pat = r"(.*\+--ro) ([A-z-\d]+)(\??)\s*(\[(.*)\])?"
        sub_tree = ""
        keys = []
        is_leaf = 1
        parent_node = ""
        for line in open("nc_tree", "r"):
            if line.startswith('module'):
                module = line.split(":")[1].strip()
            if "+" in line:
                match = re.match(pat, line)
                final_uri = ""
                level = self.get_level(len(match.group(1)))+1
                node = match.group(2)

                if match.group(2) in keys:
                    keys_dict[match.group(2)] = []
                    keys.remove(match.group(2))
                    continue

                if level >= len(list_elem):
                    list_elem.append(node)
                else:
                    list_elem[level] = node
                    list_elem = list_elem[:level+1]

                if not match.group(3):
                    is_leaf = 0
                    parent_node = "/".join(list_elem)
                    sub_tree = parent_node
                    if match.group(5):
                        keys = match.group(5).split()
                        sub_tree += "/"+node
                        
                final_uri  = "/".join(list_elem)

                if not is_leaf:
                    if len(keys):
                        sub_tree += "/"+"/".join(keys)
                        list_elem[level] = node+"/"+'/'.join(['v@l_'+x+'_v@l' for x in keys])
                    is_leaf = 1

                uri_list.append(module+final_uri)

def cmd_list_2_xml(cmd_list):
    doc = xml.dom.minidom.Document()
    el_rpc = doc.createElementNS("urn:ietf:params:xml:ns:netconf:base:1.0", "rpc")
    el_rpc.setAttribute("message-id", "101")
    el_rpc.setAttribute("xmlns", "urn:ietf:params:xml:ns:netconf:base:1.0")
    doc.appendChild(el_rpc)
    el_get = doc.createElement("get")
    el_rpc.appendChild(el_get)
    el_filter = doc.createElement("filter")
    el_filter.setAttribute("type", "subtree")
    el_get.appendChild(el_filter)
    el_main = doc.createElementNS("urn:brocade.com:mgmt:"+cmd_list[0], cmd_list[1])
    el_main.setAttribute("xmlns", "urn:brocade.com:mgmt:"+cmd_list[0])
    el_filter.appendChild(el_main)
    el_prev = el_main
    for ele in cmd_list[2:]:
        if ":" in ele:
            key, value = ele.split(':')
            el_key = doc.createElement(key)
            el_prev.appendChild(el_key)
            el_key_content = doc.createTextNode(value)
            el_key.appendChild(el_key_content)
        else:
            el_sub = doc.createElement(ele)
            el_prev.appendChild(el_sub)
            el_prev = el_sub
    return doc.toprettyxml()

def set_test_result(code):
    global ret_code
    global test_result
    ret_code = str(code)+" "+responses[code][0]
    if 200 == code:
        test_result = "PASS"
    else:
        test_result = "FAIL"

def gen_valid_combos():
    valid_combos = []
    val_list = []
    new_val_list = []
    cont = 0

    for ele in grouped_keys:
        if ele[0] in keys_dict.keys():
            val_list.append([keys_dict[x] for x in ele])
            cont = 1
    if cont:
        for ele in val_list:
            new_val_list.append(list(map(list, zip(*ele))))
        combos = list(map(list, list(zip(*new_val_list))))
        for ele in combos:
            valid_combos.append(list(itertools.chain.from_iterable(ele)))
    return valid_combos

def perform_query(cmd):
    global query
    print("Executing NetConf GET request for ", cmd)
    cmd = cmd.split("$")
    print(cmd)
    query = cmd_list_2_xml(cmd)+"]]>]]>"
    ssh_session.sendline(query)
    try:
        ssh_session.expect(['</rpc-reply>]]>]]>', pexpect.EOF])
    except pexpect.ExceptionPexpect as e:
        print("Failed to perform NetConf operation.\n")
        print("Failed Query:\n"+query+"\n")
        exit()
    output = ssh_session.before
    format_output(cmd, output)

def walk_node(cmd):
    req = ''
    cmd = "$".join(cmd.split('/'))
    if "v@l_" in cmd:
        for ele in gen_valid_combos():
            req = ""
            i = 0
            for sub_ele in cmd.split("v@l"):
                if sub_ele.startswith("_"):
                    req += sub_ele[1:-1]+":"+str(ele[i])
                    i += 1
                else:
                    req += sub_ele
            perform_query(req)
    else:
        perform_query(cmd)

def write_test_header(cmd, err_str):
    test_result = "Pass"
    outfile.write("*********************************************************************************************************************************\n")
    outfile.write("================                           =============\n")
    outfile.write("|      TEST    |                           |    "+test_result+"    |\n")
    outfile.write("================                           =============\n")
    outfile.write("*********************************************************************************************************************************\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")
    outfile.write("NETCONF URI".ljust(10)+": "+"/".join(cmd)+"\n")
    outfile.write("STATUS".ljust(10)+": "+ret_code+"\n")
    outfile.write("ERRORS".ljust(10)+": "+err_str+"\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")

def get_all_children(parent):
    for child in parent.getchildren():
        if not child.getchildren():
            tag = child.tag.split("}")[1].strip()
            outfile.write(tag.ljust(50)+' = '+child.text+"\n")
            update_keys(tag, child.text)
        else:
            tag = child.tag.split("}")[1].strip()
            outfile.write(tag.ljust(50)+"\n")
            get_all_children(child)

def format_output(cmd, buffer):
    global keys_dict
    node = "{urn:brocade.com:mgmt:"+cmd[0]+"}"+ cmd[-1]
    err_str = "None"
    data_pat = b'<data>(.*)</data>'
    data = re.search(data_pat, buffer)
    #print(buffer.decode('utf-8'))
    if not data:
        print("No response received from the device.\n")
        print("Failed Query:\n"+query+"\n")
        return

    write_test_header(cmd, err_str)
    outfile.write("Output Dump:\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")
    outfile.write(buffer.decode('utf-8'))
    outfile.write("\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")
    outfile.write("Formatted Output:\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")

    if not len(data.group(1)):
        outfile.write("Empty Response\n")
        return

    if len(buffer):
        parser = ET.XMLParser(encoding="utf-8")
        tree = ET.fromstring(data.group(1).decode('utf-8'), parser=parser)

        for child in tree:
            get_all_children(child)
        outfile.write("-----------------------------------------------------------------------------------------------\n")
        outfile.write("*********************************************************************************************************************************\n\n\n")
    elif test_result == "PASS" or len(ret_code):
        write_test_header(cmd, err_str)
    else:
        print("EXECUTION OF TEST CASES FAILED. IT MIGHT BE DUE TO THE FOLLOWING REASONS.\
               \n1.URI NOT FOUND. \
               \n2. DEVICE CRASHED.\
               \n3. CONNECTIVITY LOST.")

def is_only_list_key(ref_list, act_list):
    diff = list(set(ref_list) ^ set(act_list))

def group_keys(uri):
    #Written to track all the keys that are together and not perform permutations
    #over them.
    global grouped_keys
    grouped_keys = []
    old_group = 2
    new_group = 1
    copy = 0
    for ele in uri.split("_"):
        if ele == "v@l/v@l":
            copy = old_group
        elif ele.endswith("v@l"):
            copy = new_group
        else:
            if copy == new_group:
                grouped_keys.append([ele])
            elif copy == old_group:
                grouped_keys[-1].append(ele)
            copy = 0

def main():
    global sub_tree_node
    global param_dict
    global outfile
    sub_tree_list = []
    read_conf_file()
    if not os.path.exists(os.getcwd()+"/nc_tree"):
        create_net_conf_tree()

    if os.path.exists(param_dict["CONFIGFILE"]):
        print("Please wait...Pushing Config to the device")
        doSsh(param_dict["DEVICEUNAME"], param_dict["DEVICEPASSWD"], 22, param_dict["DEVICEIP"], param_dict["CONFIGFILE"])
    
    outputdir = param_dict["OUTPUTDIR"]
    if ("" != outputdir):
       outputdir += "/"
    file_name = "testOutput_"+strftime("%Y-%m-%d_%H-%M-%S", gmtime())
    optional_args()
    uri_dump = read_tree()
    uri_dump.print_uri()

    if not os.path.exists(outputdir+"NET_CONF_OPER_URI_DUMP"):
        out = open(outputdir+"NET_CONF_OPER_URI_DUMP", "w")
        for ele in uri_list:
            out.write(ele+"\n")
        out.close()

    if dump_gen:
        print("For URI dump refer to ", outputdir+"NET_CONF_OPER_URI_DUMP")
        print("\n*********************************************************************************************\n")
        exit()

    if sub_tree_node is None:
        #Handle to accomodate the base URI for netconf
        sub_tree_node=root_uri

    outfile = open(outputdir+file_name, "w")
    print("Executing NET CONF RPC request for ",sub_tree_node)

    init_net_conf_ssh(param_dict["NETCONFUNAME"], param_dict["NETCONFPASSWD"], '830', param_dict["DEVICEIP"], ' -s netconf')
    group_keys(sub_tree_node)
    process_keys(input_keys)
    walk_node(sub_tree_node)
    for ele in uri_list:
        if ele.startswith(sub_tree_node+"/"):
             group_keys(ele)
             walk_node(ele)

    close_net_conf_ssh()
    outfile.close()
    print("*********************************************************************************************\n")
    print("For URI dump refer to ", outputdir+"NET_CONF_OPER_URI_DUMP")
    print("Output is written to ", outputdir+file_name)
    print("*********************************************************************************************\n")

if __name__ == '__main__':
    main()
